#==========================================================================
# Promira Interface Library
#--------------------------------------------------------------------------
# Copyright (c) 2017 Total Phase, Inc.
# All rights reserved.
# www.totalphase.com
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
# - Neither the name of Total Phase, Inc. nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#==========================================================================

#==========================================================================
# CONSTANTS
#==========================================================================
VERSION = '1.0.0'


#==========================================================================
# IMPORTS
#==========================================================================
import json
import os
import sys
import threading

sys.path.insert(0, os.path.dirname(__file__) + '/ws4py.pyz')
from ws4py.client.threadedclient import WebSocketClient


#==========================================================================
# CLASS
#==========================================================================
class Promira:
    def __init__ (self, url):
        self.url = url
        self.ws  = None

        self._req_id = 1
        self._req_funcs = { }
        self._req_calls = { }

    def connect (self, session=None):
        headers = [('Cookie', 'session=' + session)] if session else None
        self.ws = WebSocketClient(self.url, headers=headers)
        self.ws.received_message = self._message
        return self.ws.connect()

    def disconnect (self):
        ws = self.ws
        self.ws = None
        return ws.close()

    def __getattr__ (self, attr):
        if attr not in self._req_funcs:
            def call (*args):
                return self._call(attr, *args)

            self._req_funcs[attr] = call

        return self._req_funcs[attr]

    def _call (self, name, *args):
        req_id = self._req_id
        self._req_id += 1

        rpc = json.dumps({
            'jsonrpc': '2.0',
            'method': name,
            'params': args,
            'id': req_id,
        })

        evt = threading.Event()

        self._req_calls[req_id] = evt
        self.ws.send(rpc)

        evt.wait()

        if evt.error != None:
            raise evt.error

        return evt.result

    def _message (self, msg):
        if msg.is_binary:
            if 'binary' in self.__class__.__dict__:
                self.binary(msg.data)
            return

        rpc = json.loads(str(msg))

        # If this is a response to a call, process the result or exception.
        if 'id' in rpc:
            evt = self._req_calls[rpc['id']]
            del self._req_calls[rpc['id']]

            if 'result' in rpc:
                evt.result = rpc['result']
                evt.error  = None
                evt.set()

            else:
                evt.error = Exception(
                    rpc['error']['code'],
                    rpc['error']['message']
                )
                evt.set()

        # Otherwise, it is a notification that ignores the result or exception.
        else:
            try:
                notify = self.__class__.__dict__.get(rpc['method'], None)
                args = rpc['params']

                if notify != None:
                    notify(self, *args)

            except Exception as ex:
                if isinstance(ex, TypeError) and \
                   str(ex).startswith(rpc['method']):
                    print(ex)

